#!/usr/bin/perl
use strict;

#--------------------------------------------------------------------
#  Apache Live Log: Monitoring multiple apache log files
#
#  Version : 0.9
#  Date    : 04 February 2009
#  Homepage: http://www.burlaca.com/2009/02/alivelog/
#  Author  : Oleg Burlaca (email: oleg@burlaca.com)
#--------------------------------------------------------------------

# one access.log file per line: title, filename, websiteUrl (without the trailing slash)

my $logs = <<__LOGS__;
Host1, /var/www/html/host1/access_log, http://www.host1.com
Host2, /var/www/html/host2/access_log, http://www.host2.com
__LOGS__

our $cfg = {
    # the maximum Nr of lines to process from the log file
    # Note: a line that doesn't satisfy search criteria is counted 
    # as a processed line
    maxLogLines => 1000,   


    # the location of 'GeoLiteCity.dat' file
    geoIPdat	=> '/usr/local/share/GeoIP/GeoLiteCity.dat',


    # set useGeoIP => 0 if you don't need to see the location of the user
    # you can also comment the "use Geo::IP ()" line below
    useGeoIP	=> 1, 
    

    # These files will be ignored
    # if their size is smaller than ignoreFilesMinSize 
    ignoreFiles => 'txt,css,js,jpg,png,gif,ico',     


    # the filesize in bytes.
    # a zero value means that all ignoreFiles will be skipped
    ignoreFilesMinSize => 40000,		     
    

    # a comma separated list of IP's to skip: ignoreIps => '192.168.2.1,192.168.2.100'
    ignoreIPs	=> '',
    

    # a list of bots, add yours.
    # requests made by bots are colored in grey and the Referrer column shows the User-Agent
    botAgent => 'appie,architext,jeeves,bjaaland,contentmatch,ferret,gulliver,harvest,htdig,' .
		'feedburner,feedvalidator,lycos,moget,netcraft,ariadne,auresys,infoseek,w3index' .
		'linkwalker,followsite',
    
    logsRaw => $logs,
};
#------------------------------------------------------------------------------------------------------

use JSON ();
use Geo::IP ();
use File::ReadBackwards ();
use Date::Parse ();


print "Content-type: text/html\n\n";
our $params = &getCGIParams();

#$params->{cmd} = 'getApacheTail';
#$params->{logId} = 2;
#$params->{interval} = '30m';

&init();

if ($params->{cmd} eq 'getLogs') {
    print &getLogFiles();
} elsif ($params->{cmd} eq 'getApacheTail') {
    print &getApacheTail($params->{logId});
} else {
    print "ALiveLog: No command specified!\n";
}


#---------------------------------------------------------------------------------------------------
sub getApacheTail {
    my ($logId, $interval) = @_;
    my $log = $cfg->{logs}[$logId];
    my $siteUrl = $log->{url};
    
    &setupInterval();
    # print "lastTime: ", $cfg->{_limitTime}, "\n";
    
    my $h = {scannedLines => 0, totalBytes => 0, totalHits => 0, totalVisits => 0};
    my $visits = {};
    my @sortedIp = ();
    
    # opening access.log file
    my $filename = $cfg->{logs}[$logId]{filename};
    my $bw = File::ReadBackwards->new($filename) or
                        die "can't read '$filename' $!" ;
    # reading the access.log file
    while( defined( my $l = $bw->readline ) and $h->{scannedLines} < $cfg->{maxLogLines}) {
	$h->{scannedLines}++;
	my $hit = &parseLogLine($l);
	next if &ignoreHit($hit);
	last if ($cfg->{_intervalType} eq 'm' and &getSeconds($hit->{dt}, $hit->{tm}) < $cfg->{_limitTime});
	
	$h->{totalHits}++;
	
	# @sortedIp array holds IP's in the order of appearence in the log file
	if (! exists $visits->{$hit->{host}}) {
	    push(@sortedIp, $hit->{host});
	    $visits->{$hit->{host}} = {};
	    $h->{totalVisits}++;
	}
	
	my $v = $visits->{$hit->{host}};
	
	# create the hits array for a new IP
	$v->{hits} = [] if (! exists $v->{hits});

	&processHit($v, $hit, {siteUrl => $siteUrl});
	$h->{_firstHitTime} = $v->{_firstHit};
	
	# check if the restrictions are satisfied
	last if &outOfInterval($h);
    }
    
    $h->{timeInterval} = $h->{_firstHitTime} ? $cfg->{_nowTime} - $h->{_firstHitTime}
					     : 0;
    
    # create the JSON response
    $h->{visits} = [];
    foreach my $ip (@sortedIp)  {
	my $d = $visits->{$ip};
	$d->{ip} = $ip;
	
	$d->{fp} = $d->{hits}[0]{u};
	$d->{lp} = $d->{hits}[-1]{u} if @{$d->{hits}} > 1;
	
	$d->{dt} = $d->{hits}[0]{dt};
	$d->{tm} = $d->{hits}[0]{tm};
	
	$h->{totalBytes} += $d->{sz};
	
	&appendGeoInfo($d);
	push(@{$h->{visits}}, $visits->{$ip});
    }
    
    return JSON::to_json($h);
    #return Dumper($h->{visits});
}
#---------------------------------------------------------------------------------------------------
sub outOfInterval {
    my ($h) = @_;
    my $t = $cfg->{_intervalType};
#    if ($t eq 'm') {
#	return $h->{_firstHitTime} < $cfg->{_limitTime};
#    } 
    if ($t eq 'v') {
	return $h->{totalVisits} == $cfg->{_intervalValue};
    } elsif ($t eq 'h') {
	return $h->{totalHits} == $cfg->{_intervalValue};
    }
    return 0;
}
#---------------------------------------------------------------------------------------------------
sub processHit {
    my ($v, $hit, $p) = @_;
    
    my $sec = &getSeconds($hit->{dt}, $hit->{tm});
    $v->{_firstHit} = $sec;
    $v->{_lastHit} = $sec if (! exists $v->{_lastHit});
    $v->{sz} += $hit->{bytes};
    $v->{t}++;
    
    # The referrer is the last referrer that is not on the domain
    $v->{refer} = $hit->{refer} if (! exists $v->{refer} and $hit->{refer} ne '-' and index($hit->{refer}, $p->{siteUrl}) == -1);
    
    if (! exists $v->{agent}) {
	$v->{agent} = $hit->{agent};
	&agentBotCheck($v);
    }    

    my $h = {dt => $hit->{dt}, 		# date
	  tm => $hit->{tm}, 		# time
	  u => $hit->{file}, 		# requested page uri
    };
    $h->{rf} = $hit->{refer} if $hit->{refer} ne '-';          # referrer
    

    $h->{c} = $hit->{code};
    if ($hit->{code} > 400) {           
	# $h->{c} = $hit->{code};
	if ($hit->{code} >= 500) {
	    $v->{errS}++;     # Server Error 5xx
	} else {
	    $v->{errC}++;     # Client Error 4xx
	}
	$v->{err} = $v->{errS} + $v->{errC};  # the total number of errors
    }    
    
    push(@{$v->{hits}}, $h);
	  
}
#-----------------------------------------------------------------------------------------------
sub agentBotCheck {
    my ($v) = @_;
    my $t = lc($v->{agent});
    
    # check if: Googlebot, MSNBot, Yahoo or FeedBurner
    if (index($t, 'googlebot') != -1) {
	$v->{bot} = 'GoogleBot';
    } elsif (index($t, 'msnbot') != -1) {
	$v->{bot} = 'MsnBot';
    } elsif (index($t, 'yahoo! slurp') != -1) {
	$v->{bot} = 'YahooBot';
    } elsif (index($t, 'feedburner') != -1) {
	$v->{bot} = 'FeedBurner';
    }
    
    return if $v->{bot};
    
    # check the rest, we don't need the name of the robot, just True/False
    foreach (@{$cfg->{_botAgents}}) {
	if (index($t, $_) != -1) {
	    $v->{bot} = 1;
	    return;
	}
    }
}
#---------------------------------------------------------------------------------------------------
#  dt = 30/Jan/2009
#  tm = 04:46:54
sub getSeconds {
    my ($dt, $tm) = @_;
    return Date::Parse::str2time("$dt $tm");
}
#---------------------------------------------------------------------------------------------------
sub ignoreHit {
    my ($hit) = @_;

    return 1 if (exists $cfg->{_ignoreIPhash}{$hit->{host}});

    my $t = $hit->{uri};
    return 0 if chop($t) eq '/';   # a /dir1/dir2/ uri 
    
    my $ext = substr($hit->{uri}, rindex($hit->{uri}, '.')+1, length($hit->{uri}));
    # $hit->{uri} =~ m/\.(.+)$/;
    $ext = ',' . $ext . ',';
    
    return (index($cfg->{ignoreFiles}, $ext) != -1 and $hit->{bytes} < $cfg->{ignoreFilesMinSize});
}
#---------------------------------------------------------------------------------------------------
sub init {
    my $t = $cfg->{logsRaw};
    $cfg->{logs} = [];
    foreach my $l (split(/\n/,$cfg->{logsRaw})) {
	my @a = split(/,\s*/, $l);
	push(@{$cfg->{logs}}, {title => $a[0], filename => $a[1], url => $a[2]});
    }
    $cfg->{ignoreFiles} =~ s/\s+//g;
    $cfg->{ignoreFiles} = ',' . $cfg->{ignoreFiles} . ',';
    my @a = split(/,/, $cfg->{botAgent});
    $cfg->{_botAgents} = \@a;
    
    $cfg->{_ignoreIPhash} = {};
    if ($cfg->{ignoreIPs}) {
	foreach (split(/,/, $cfg->{ignoreIPs})) {$cfg->{_ignoreIPhash}{$_} = 1 }
    }
    
    if ($cfg->{useGeoIP}) {
        my $gi = Geo::IP->open($cfg->{geoIPdat}, Geo::IP::GEOIP_STANDARD)
                    or die "Can't open GeoIP file: ", $!;
        $cfg->{gi} = $gi;
    }
}
#---------------------------------------------------------------------------------------------------
sub getLogFiles {
    my @a = ();
    foreach my $d (@{$cfg->{logs}}) {
	push @a, {title => $d->{title}, url => $d->{url}};
    }
    return JSON::to_json(\@a);
}

#---------------------------------------------------------------------------------------------------
#  this function was copied from Parse-AccessLogEntry CPAN module
#
sub parseLogLine {
    my $Line = shift;
    my $Ref;
    my $Rest;
    my $R2;
    ($Ref->{host},$Ref->{user},$Ref->{date},$Rest)= $Line=~m,^([^\s]+)\s+-\s+([^ ]+)\s+\[(.*?)\]\s+(.*),;
    my @Dsplit=split(/\s+/,$Ref->{date});
    $Ref->{diffgmt}=$Dsplit[1];
    my @Ds2=split(/\:/,$Dsplit[0],2);
    $Ref->{dt}=$Ds2[0];
    $Ref->{tm}=$Ds2[1];
    if ($Rest)
    {
	($Ref->{rtype},$Ref->{file},$Ref->{proto},$Ref->{code},$Ref->{bytes},$R2)=split(/\s/,$Rest,6);
	$Ref->{rtype}=~tr/\"//d;
	$Ref->{proto}=~tr/\"//d;
	if ($R2)
	{
	    my @Split=split(/\"/,$R2);
	    $Ref->{refer}=$Split[1];
	    $Ref->{agent}=$Split[3];
        }
    }
    
    
    # post processing
    if (index($Ref->{file}, '?') != -1) {
	$Ref->{uri} = substr($Ref->{file}, 0, index($Ref->{file}, '?'));
    } else {
	$Ref->{uri} = $Ref->{file};
    } 
    
    return $Ref;
}
#---------------------------------------------------------------------------------------------------
sub appendGeoInfo {
    my ($d) = @_;
    
    if ($cfg->{useGeoIP}) {
        my $gi = $cfg->{gi};
        my $r = $gi->record_by_name($d->{ip});
    
        $d->{cnCode} = $r->country_code;
        $d->{cnName} = $r->country_name;
        $d->{region} = $r->region_name;
        $d->{city} = $r->city;
        
        # Geo coords will be used when GoogleMaps will be added to the frontend
        # $d->{geoLt} = $r->latitude;
        # $d->{geoLg} = $r->longitude;
    } else {
	$d->{cnCode} = '';
        $d->{cnName} = '';
        $d->{region} = '';
        $d->{city} = '';
    }
}                                                               
#---------------------------------------------------------------------------------------------------
sub setupInterval {
    $cfg->{_intervalValue} = $params->{interval};
    $cfg->{_intervalType} = chop($cfg->{_intervalValue});
    
    $cfg->{_nowTime} = time();
    
    if ($cfg->{_intervalType} eq 'm') {
        $cfg->{_limitTime} =  $cfg->{_nowTime} - 60*$cfg->{_intervalValue};
    }
}
#---------------------------------------------------------------------------------------------------
sub getCGIParams {
    my $line;
    
    if ($ENV{'REQUEST_METHOD'} eq "POST") {
        read(STDIN, $line, $ENV{'CONTENT_LENGTH'});
    } else {
        $line = $ENV{'QUERY_STRING'};
    }

    my (@pairs) = split(/&/, $line);
    my ($name, $value, %F);
        
    foreach (@pairs) {
        ($name, $value) = split(/=/);
        $value =~ tr/+/ /;
        $value =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;

        if (! exists $F{$name}) {
            $F{$name} = $value;
        } elsif (exists $F{$name} and ref($F{$name}) ne 'ARRAY') {
            my $prev_value = $F{$name};
            delete $F{$name};
            $F{$name} = [ $prev_value, $value ];
        } else { push @{ $F{$name} }, $value }
    }
                                        
    return \%F;
}
#---------------------------------------------------------------------------------------------------                                                                                                                                                            

